@@ -194,14 +194,18 @@ class ElucidatedDiffusion(nn.Module):
                 images_next = images_hat + 0.5 * (sigma_next - sigma_hat) * (denoised_over_sigma + denoised_prime_over_sigma)
 
             images = images_next
-            x_start = model_output
+            x_start = model_output_next if sigma_next != 0 else model_output
 
         images = images.clamp(-1., 1.)
         return unnormalize_to_zero_to_one(images)
 
-    # thanks to Katherine Crowson (https://github.com/crowsonkb) for figuring it all out!
     @torch.no_grad()
     def sample_using_dpmpp(self, batch_size = 16, num_sample_steps = None):
+        """
+        thanks to Katherine Crowson (https://github.com/crowsonkb) for figuring it all out!
+        https://arxiv.org/abs/2211.01095
+        """
+
         device, num_sample_steps = self.device, default(num_sample_steps, self.num_sample_steps)
 
         sigmas = self.sample_schedule(num_sample_steps)
